"""
Representation for a set of rar files, but more generically any split
archives.
"""

#TODO:
# - revise the notion of complete files in rarset
# - make rarset more generic and the patterns customizable

import os
import re

from itertools import groupby

__all__ = ['find_largest_rarset', 'is_complete']

# remove numeric suffixes that nget puts on for incomplete files
STRIP = re.compile(r'\.\d{6,}\.\d{6,}$')

# split out the base into $1 and the numeric part into named groups
# depending on what naming convention is used.  Does not currently support
# more than 100ish volumes (r99 + 1 = s00)
RAR = re.compile(r'^(.+?)\.(?:(?P<bare>\d{3,})|part(?P<pre>\d+)\.rar|r(?P<ext>ar|\d{2}))$')

def partition(files):
    def _grouper(fn):
        f2 = STRIP.sub('', fn)
        m = RAR.match(f2)
        if not m:
            return fn
        else:
            return m.group(1)
    return map(RarSet, groupby(files, _grouper))

def fn_to_index(fn):
    """
    Convert a rar filename into its volume index.

    >>> map(fn_to_index, ['x.part0%d.rar' % i for i in [1,2,3]])
    [0, 1, 2]
    >>> map(fn_to_index, ['x.rar', 'x.r00', 'x.r01'])
    [0, 1, 2]
    >>> map(fn_to_index, ['x.001', 'x.002', 'x.003'])
    [0, 1, 2]
    """
    f2 = STRIP.sub('', fn)
    m = RAR.match(f2)
    if not m:
        # This is kind of annoying.
        return 9999
    if m.group('ext'):
        if m.group('ext') == 'ar':
            n = 0
        else:
            n = int(m.group('ext')) + 1
    else:
        n = int(m.group('pre') or m.group('bare')) - 1
    return n


def is_complete(fn):
    """
    Whether the `STRIP` re strips anything off

    >>> is_complete('x.rar.123123123.123123123')
    False
    >>> is_complete('x.rar')
    True
    """
    f2 = STRIP.sub('', fn)
    return fn == f2


def annotate(list):
    """
    For a list of filenames, turn it into a list of tuples containing
    the volume index of each, starting at 0.

    >>> annotate(['file.part01.rar', 'file.part03.rar'])
    [(0, 'file.part01.rar'), (2, 'file.part03.rar')]
    >>> annotate(['file.001', 'file.003'])
    [(0, 'file.001'), (2, 'file.003')]
    >>> annotate(['file.rar.123123123.123123123'])
    [(0, 'file.rar.123123123.123123123')]
    """
    return [(fn_to_index(f), f) for f in list]


class RarSet(object):
    """
    Represents a set of archives (designed for rars)
    """
    def __init__(self, files):
        # files[0] is the group name, files[1] is a grouper object
        self.base_name = files[0]
        self.files = sorted(annotate(list(files[1])))
        files = list(files[1])
        
        # This presumes that files are either the proper length or too
        # short, never too long.  Based on this assumption, find the maximum
        # size...
        self.volume_size = 0
        for i, f in self.files:
            size = os.stat(f)[6]
            if size > self.volume_size:
                self.volume_size = size

    def first_volume(self):
        return self.files[0]
    def best_bets(self):
        """
        Returns something like [(0, ('file01.rar', 15000)), (1, ('file02.rar', 15000))]
        """
        bets = {}
        for i, f in self.files:
            size = os.stat(f)[6]
            # First time we see files, add them regardless of whether they
            # appear to be incomplete.
            if i not in bets or not STRIP.search(f):
                bets[i] = (f, size)
            elif size > bets[i][1]: # Larger than the current one
                bets[i] = (f, size)
        return sorted(bets.items())

    def __repr__(self):
        return "<Rarset for base %s with %d files>" % \
               (self.base_name, len(self.files))

def find_largest_rarset(filenames):
    sets = partition(filenames)
    biggest = sorted(sets, key=lambda x: len(x.files))
    return biggest[-1]

if __name__ == '__main__':
    from doctest import testmod
    testmod()
