"""
Module for resyncing data based on cryptographic hashes.  See docstring for
attempt_sync for examples.
"""

SEARCH_FORWARD = 1
SEARCH_BACKWARD = 2
SEARCH_BOTH = 1 + 2

import sha

class SyncError(Exception):
    pass

hello_hash = '\xaa\xf4\xc6\x1d\xdc\xc5\xe8\xa2\xda\xbe\xde\x0f;H,\xd9\xae\xa9CM'

def attempt_sync(data, initial_pos, hash, blocksize, prev_data=None,
                 direction=SEARCH_BOTH, distance=200, hashfunc=sha.new):
    """
    Find a substring of `blocksize` from `data` that matches the provided
    sha `hash`.  Initial guess will be `initial_pos` and both directions
    will be searched unless specified otherwise.

    If `prev_data` is provided, it forms the first part of the block, and
    the search will only look for `blocksize` - `len(prev_data)` bytes when
    attempting to sync.

    `distance` is the number of bytes to search, in the specified
    directions.

    >>> attempt_sync('hello', 0, hello_hash, 5)
    0
    >>> attempt_sync('ahello', 0, hello_hash, 5)
    1
    >>> attempt_sync('ello', 0, hello_hash, 5, 'h')
    0
    >>> attempt_sync('___ello', 0, hello_hash, 5, 'h')
    3

    Example of backward sync

    >>> attempt_sync('hello___', 3, hello_hash, 5)
    0
    
    And it throws errors

    >>> attempt_sync('abcde', 0, hello_hash, 5)
    Traceback (most recent call last):
    ...
    SyncError
    >>> attempt_sync('a', 0, hello_hash, 5)
    Traceback (most recent call last):
    ...
    SyncError
    """
    
    prev_data = prev_data or ''

    s = hashfunc(prev_data) # hash so far
    l = len(prev_data) # previous length
    t = blocksize - l  # remaining bytes
    m = len(data)      # length of the data

    assert t > 0
    #TODO: additional check for whether it's possible to match based on
    #      lengths, to short-circuit the loop.

    # Perform a bidirectional search based on the location
    for i in range(0, distance + 1):

        # Try forward
        forward_pos = initial_pos + i
        if m - forward_pos >= t:
            s2 = s.copy()
            s2.update(data[forward_pos:forward_pos+t])
            if s2.digest() == hash:
                return forward_pos

        # Try backward
        backward_pos = initial_pos - i
        if m - backward_pos >= t:
            s2 = s.copy()
            s2.update(data[backward_pos:backward_pos+t])
            if s2.digest() == hash:
                return backward_pos

    raise SyncError()

if __name__ == '__main__':
    import doctest
    doctest.testmod()

